/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.Iterator;
import java.util.List;

import com.ibm.etools.c.CArray;
import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDatatype;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CEnumerator;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CPointer;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedElement;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.datatypes.CBitField;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.typedescriptor.AccessorValue;
import com.ibm.etools.typedescriptor.AddrUnitValue;
import com.ibm.etools.typedescriptor.AddressMode;
import com.ibm.etools.typedescriptor.AddressTD;
import com.ibm.etools.typedescriptor.AlignType;
import com.ibm.etools.typedescriptor.ArrayTD;
import com.ibm.etools.typedescriptor.BaseTDType;
import com.ibm.etools.typedescriptor.BinaryTD;
import com.ibm.etools.typedescriptor.FloatTD;
import com.ibm.etools.typedescriptor.FloatValue;
import com.ibm.etools.typedescriptor.InstanceTDBase;
import com.ibm.etools.typedescriptor.IntegerTD;
import com.ibm.etools.typedescriptor.LengthEncodingValue;
import com.ibm.etools.typedescriptor.PlatformCompilerInfo;
import com.ibm.etools.typedescriptor.SignCodingValue;
import com.ibm.etools.typedescriptor.SimpleInstanceTD;
import com.ibm.etools.typedescriptor.StringTD;
import com.ibm.etools.typedescriptor.TypeDescriptorFactory;
import com.ibm.etools.typedescriptor.TypeDescriptorPackage;

/**
 * CTypeDescriptorBuilder is used to create TypeDescriptor instance and type
 * objects.
 */
public class CTypeDescriptorBuilder {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private static final int SIZE_OF_ENUM_COMPACT_VALUE = 5; // @P3

    private static CTypeDescriptorBuilder _instance = new CTypeDescriptorBuilder();

    private static TypeDescriptorPackage _tdPkg;

    private static TypeDescriptorFactory _tdFactory;

    private static PlatformCompilerInfo _platformCompilerInfo = null;

    private final static String IEEE_EXTENDED_INTEL = "IEEE Extended INTEL";

    private final static String IEEE_EXTENDED_AIX = "IEEE Extended AIX";

    private final static String IEEE_EXTENDED_OS390 = "IEEE Extended OS/390";

    private final static String IEEE_EXTENDED_AS400 = "IEEE Extended AS400";

    private final static String IEEE_NON_EXTENDED = "IEEE Non-Extended";

    private final static String IBM_390_HEX = "IBM 390 Hexadecimal";

    private final static String IBM_400_HEX = "IBM 400 Hexadecimal";

    private final static String COMPILER_UNSPECIFIED = "(Unspecified)";

    /**
     * Create a CTypeDescriptorBuilder object
     */
    private CTypeDescriptorBuilder() {
        //_tdPkg =
        // (TypeDescriptorPackage)RefRegister.getPackage(TypeDescriptorPackage.packageURI);
        _tdPkg = (TypeDescriptorPackage) org.eclipse.emf.ecore.EPackage.Registry.INSTANCE.getEPackage(TypeDescriptorPackage.eNS_URI);
        _tdFactory = _tdPkg.getTypeDescriptorFactory();
    }

    /**
     * Return the only instance of CTypeDescriptorBuilder
     */
    public static CTypeDescriptorBuilder instance() {
        return _instance;
    }

    /**
     * Create a PlatformCompilerInfo object
     */
    public static void createPlatformCompilerInfo() {
        _platformCompilerInfo = _tdFactory.createPlatformCompilerInfo();
        _platformCompilerInfo.setLanguage("C");
        _platformCompilerInfo.setCompilerFlags(CPreferenceStore.getCompilerFlags());
        _platformCompilerInfo.setDefaultBigEndian(new Boolean(CPreferenceStore.isBigEndian()));

        // Set the optional attributes
        String work = null;
        work = CPreferenceStore.getCompilerName();
        if (work != null && !work.equals(COMPILER_UNSPECIFIED))
            _platformCompilerInfo.setCompilerName(work);

        if ((work = CPreferenceStore.getOperatingSystem()) != null)
            _platformCompilerInfo.setOperatingSystem(work);

        if ((work = CPreferenceStore.getHardwarePlatform()) != null)
            _platformCompilerInfo.setHardwarePlatform(work);

        if ((work = CPreferenceStore.getCompilerVersion()) != null)
            _platformCompilerInfo.setCompilerVersion(work);

        if ((work = CPreferenceStore.getOSVersion()) != null)
            _platformCompilerInfo.setOSVersion(work);

        if ((work = CPreferenceStore.getFloatingPointFormat()) != null)
            _platformCompilerInfo.setDefaultFloatType(mapFloatValue(work));

        if ((work = CPreferenceStore.getCodePage()) != null)
            _platformCompilerInfo.setDefaultCodepage(work);

        if (CPreferenceStore.is64bitAddress())
            _platformCompilerInfo.setDefaultAddressSize(AddressMode.MODE64_LITERAL);
        else
            _platformCompilerInfo.setDefaultAddressSize(AddressMode.MODE32_LITERAL);
    }

    /**
     * Create a TypeDescriptor instance object. If the isAggregate flag is set,
     * create an AggregateInstanceTD; otherwise create a SimpleInstanceTD.
     */
    public static InstanceTDBase createInstance(boolean isAggregate) {
        InstanceTDBase instance = null;
        if (isAggregate)
            instance = _tdFactory.createAggregateInstanceTD();
        else
            instance = _tdFactory.createSimpleInstanceTD();

        instance.setAccessor(AccessorValue.READ_WRITE_LITERAL);
        instance.setAttributeInBit(Boolean.FALSE);
        instance.setPlatformInfo(getPlatformCompilerInfo());
        return instance;
    }

    /**
     * Return a clone of the given Type Descriptor instance
     */
    public static InstanceTDBase clonedInstance(InstanceTDBase instance) {
        InstanceTDBase clonedInstance = null;
        // Create a simple or aggregate instance depending on the type of the
        // original instance.
        if (instance instanceof SimpleInstanceTD) {
            SimpleInstanceTD simpleInstance = _tdFactory.createSimpleInstanceTD();
            simpleInstance.setSharedType(((SimpleInstanceTD) instance).getSharedType());
            clonedInstance = simpleInstance;
        } else
            clonedInstance = _tdFactory.createAggregateInstanceTD();

        // Clone the array descriptor if it is not empty.
        List arrayDesc = instance.getArrayDescr();
        if (!arrayDesc.isEmpty()) {
            for (int i = 0; i < arrayDesc.size(); i++) {
                Object listElement = arrayDesc.get(i);
                if (listElement instanceof ArrayTD) {
                    ArrayTD clonedArrayTD = clonedArrayTD((ArrayTD) listElement);
                    clonedInstance.getArrayDescr().add(clonedArrayTD);
                }
            }
        }

        clonedInstance.setOffset(instance.getOffset());
        clonedInstance.setContentSize(instance.getContentSize());
        clonedInstance.setSize(instance.getSize());
        clonedInstance.setAccessor(instance.getAccessor());
        clonedInstance.setAttributeInBit(instance.getAttributeInBit());
        clonedInstance.setPlatformInfo(instance.getPlatformInfo());
        return clonedInstance;
    }

    /**
     * Return a clone of an ArrayTD object
     */
    private static ArrayTD clonedArrayTD(ArrayTD arrayTD) {
        ArrayTD clonedArrayTD = _tdFactory.createArrayTD();
        clonedArrayTD.setStride(arrayTD.getStride());
        clonedArrayTD.setStrideInBit(arrayTD.getStrideInBit());
        clonedArrayTD.setUpperBound(arrayTD.getUpperBound());
        clonedArrayTD.setLowerBound(arrayTD.getLowerBound());
        clonedArrayTD.setAlignmentKind(arrayTD.getAlignmentKind());
        return clonedArrayTD;
    }

    /**
     * Create a TypeDescriptor instance for a given typed element.
     */
    public static InstanceTDBase createInstanceForTypedElement(CTypedElement element) throws CException {
        return createInstanceForField(CQuery.typeOf(element));
    }

    /**
     * Create a TypeDescriptor instance for a given field.
     */
    public static InstanceTDBase createInstanceForField(CClassifier element) throws CException {

        InstanceTDBase instance = createInstance(CQuery.isAggregate(element));
        CClassifier baseType = setArrayDescForInstance(instance, element);
        // The offset and size are in bits for a bitfield.
        if (baseType instanceof CBitField)
            instance.setAttributeInBit(Boolean.TRUE);

        // Set the type for a simple instance
        if (instance instanceof SimpleInstanceTD) {
            SimpleInstanceTD simpleInstance = (SimpleInstanceTD) instance;
            BaseTDType sharedType = null;
            try {
                sharedType = getTDType(baseType);
            } catch (Exception e) {
                CResource.writeLog(CTypeDescriptorBuilder.instance(), "createInstanceForTypedElement", e);
            }
            if (sharedType != null)
                simpleInstance.setSharedType(sharedType);
        }

        return instance;
    }

    /**
     * Create a Type Descriptor instance for a struct/union field. The size and
     * offset are not set for now. They will be set later.
     */
    public static InstanceTDBase createInstanceForField(CField afield) throws CException {
        return createInstanceForTypedElement(afield);
    }

    /**
     * Create a TypeDescriptor instance for a global variable or a function
     * parameter. Set the size and offset here. The offset is always 0.
     */
    public static InstanceTDBase createInstanceForVariable(CTypedElement avar) throws CException {
        InstanceTDBase instance = createInstanceForTypedElement(avar);
        instance.setOffset("0");
        if (instance.getArrayDescr().size() == 0) {
            int size = CTypeDescriptorHelper.sizeof(avar);
            instance.setSize(String.valueOf(size));
            instance.setContentSize(instance.getSize());
        }
        return instance;
    }

    /**
     * Set the array descriptor for an instance and return the base type of the
     * instance.
     * 
     * @param instance
     *            The TypeDescriptor instance.
     * @param type
     *            The C model type of the typed element
     */
    private static CClassifier setArrayDescForInstance(InstanceTDBase instance, CClassifier type) throws CException {
        if (type == null)
            return null;

        CClassifier baseType = type;
        // If it is a typedef, retrieve the original base type.
        if (type instanceof CTypedef)
            baseType = CQuery.typedefBaseType((CTypedef) type);

        // Create an ArrayTD object for each array dimension. char and wchar_t
        // arrays are ignored.
        while (baseType instanceof CArray) {
            CArray myArray = (CArray) baseType;
            Object referenceType = myArray.getDerives();
            int dimension = myArray.getDimension().intValue();
            // Arrays of char and wchar_t are represented by StringTD. Do not
            // create ArrayTD objects for them.
            if (CQuery.isWcharArray(myArray) || CQuery.isCharArray(myArray) || (CQuery.isUnsignedShortArray(myArray) && CPreferenceStore.isTreatUShortArrayAsDBCS())) //@P1A
                break;
            else if (referenceType instanceof CClassifier) {
                int elementSize = CTypeDescriptorHelper.sizeof((CClassifier) referenceType);
                int alignment = CTypeDescriptorHelper.alignmentOf((CClassifier) referenceType);
                ArrayTD arrayTD = createArrayType(String.valueOf(elementSize), alignment, String.valueOf(dimension), "1");
                instance.getArrayDescr().add(arrayTD);
                baseType = (CClassifier) referenceType;

                if (baseType instanceof CTypedef)
                    baseType = CQuery.typedefBaseType((CTypedef) baseType);
            }
        }

        // Set the size and offset now if the array descritor list is not empty.
        // The instance size is set to the size of one array element.
        if (instance.getArrayDescr().size() > 0) {
            int size = CTypeDescriptorHelper.sizeof(type); //@P2C
            instance.setSize(String.valueOf(size));
            instance.setContentSize(instance.getSize());
        }
        return baseType;
    }

    /**
     * Create a IntegerTD object that represents an integral type (int, long,
     * char, etc.)
     * 
     * @return The IntegerTD object that represents the integral type
     * @param width
     *            The width of the integral type
     * @param isSigned
     *            Is it a signed integral?
     */
    public static IntegerTD createIntegralType(int width, boolean isSigned) {

        IntegerTD int_type = _tdFactory.createIntegerTD();
        // Set the address and base unit attributes
        int_type.setAddrUnit(AddrUnitValue.BYTE_LITERAL);
        int_type.setWidth(width);
        int_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));
        int_type.setBase(2);
        int_type.setBaseWidth(1);
        int_type.setBaseInAddr(8);
        int_type.setBaseUnits(8 * width);

        // Set the alignment
        setBaseTDTypeAlignment(int_type, width);

        // Set the signed attributes
        if (isSigned) {
            int_type.setSigned(Boolean.TRUE);
            int_type.setSignCoding(SignCodingValue.TWOS_COMPLEMENT_LITERAL);
        } else {
            int_type.setSigned(Boolean.FALSE);
            int_type.setSignCoding(SignCodingValue.UNSIGNED_BINARY_LITERAL);
        }

        return int_type;
    }

    /**
     * Create a IntegerTD object to represent a bitfield. The address unit is in
     * bits for a bitfield.
     */
    public static IntegerTD createIntegralTypeForBitField(int size) {

        IntegerTD int_type = _tdFactory.createIntegerTD();
        // Set the address and base unit attributes
        int_type.setAddrUnit(AddrUnitValue.BIT_LITERAL);
        int_type.setWidth(size);
        int_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));
        int_type.setBase(2);
        int_type.setBaseWidth(1);
        int_type.setBaseInAddr(1);
        int_type.setBaseUnits(size);

        // Set the alignment
        int_type.setAlignment(AlignType.BYTE_LITERAL);

        // set the bitfield to be unsigned
        int_type.setSigned(Boolean.FALSE);
        int_type.setSignCoding(SignCodingValue.UNSIGNED_BINARY_LITERAL);

        return int_type;
    }

    /**
     * Create a BinaryTD object to represent a bitfield. The address unit is in
     * bits for a bitfield.
     */
    public static BinaryTD createBinaryTypeForBitField(int size) {

        BinaryTD binary_type = _tdFactory.createBinaryTD();
        // Set the address and base unit attributes
        binary_type.setAddrUnit(AddrUnitValue.BIT_LITERAL);
        binary_type.setWidth(size);
        binary_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));

        // Set the alignment
        binary_type.setAlignment(AlignType.BYTE_LITERAL);

        return binary_type;
    }

    /**
     * Create a FloatTD object that represents a floating point type (float,
     * double, etc.)
     */
    public static FloatTD createFloatingType(int width) {
        FloatTD float_type = _tdFactory.createFloatTD();
        float_type.setAddrUnit(AddrUnitValue.BYTE_LITERAL);
        float_type.setWidth(width);

        setBaseTDTypeAlignment(float_type, width);
        String floatType = CPreferenceStore.getFloatingPointFormat();
        float_type.setFloatType(mapFloatValue(floatType));
        float_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));

        return float_type;
    }

    /**
     * Create a StringTD object that represents a string. Char, wchar_t and char
     * array are represented by StringTD in C.
     * 
     * @return The StringTD object that represents the string type
     * @param width
     *            The width of the string
     * @param charSize
     *            The character size
     * @param isFixedLength
     *            Is this a fixed length or null-terminated string?
     */
    public static StringTD createStringType(int width, int charSize, boolean isFixedLength) {
        StringTD string_type = _tdFactory.createStringTD();
        string_type.setAddrUnit(AddrUnitValue.BYTE_LITERAL);
        string_type.setWidth(width);
        string_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));
        string_type.setCharacterSize(charSize);
        string_type.setCodepage(getPlatformCompilerInfo().getDefaultCodepage());

        setBaseTDTypeAlignment(string_type, charSize);

        // Set the length encoding attribute

        if (isFixedLength) {
            string_type.setLengthEncoding(LengthEncodingValue.FIXED_LENGTH_LITERAL);
        } else if (CPreferenceStore.isStringNullTerminated()) {
            string_type.setLengthEncoding(LengthEncodingValue.NULL_TERMINATED_LITERAL);
        } else {
            string_type.setLengthEncoding(LengthEncodingValue.FIXED_LENGTH_LITERAL);
            String work = null;
            if ((work = CPreferenceStore.getPaddingCharacter()) != null)
                string_type.setPaddingCharacter(work);
            else
                string_type.setPaddingCharacter(" ");
        }

        return string_type;
    }

    /**
     * Create an ArrayTD object for an array type
     * 
     * @return An ArrayTD object that represents the given array type
     * @param stride
     *            The stride of the array (the size of one array element in our
     *            case)
     * @param alignment
     *            The alignment of the array element
     * @param upperBound
     *            The upper bound of the array (the number of elements in our
     *            case)
     * @param lowerBound
     *            The lower bound of the array (1 in our case)
     */
    public static ArrayTD createArrayType(String stride, int alignment, String upperBound, String lowerBound) {
        ArrayTD array_type = _tdFactory.createArrayTD();
        array_type.setStride(stride);
        array_type.setStrideInBit(Boolean.FALSE);
        array_type.setUpperBound(upperBound);
        array_type.setLowerBound(lowerBound);

        // Set the alignment
        switch (alignment) {
        case 1:
            array_type.setAlignmentKind(AlignType.BYTE_LITERAL);
            break;
        case 2:
            array_type.setAlignmentKind(AlignType.HALFWORD_LITERAL);
            break;
        case 4:
            array_type.setAlignmentKind(AlignType.WORD_LITERAL);
            break;
        case 8:
            array_type.setAlignmentKind(AlignType.DOUBLEWORD_LITERAL);
            break;
        default:
            array_type.setAlignmentKind(AlignType.BYTE_LITERAL);
            break;
        }

        return array_type;
    }

    /**
     * Create an AddressTD object that represents a pointer type
     */
    public static AddressTD createAddressType() {
        AddressTD addr_type = _tdFactory.createAddressTD();
        addr_type.setAddrUnit(AddrUnitValue.BYTE_LITERAL);

        // Address is 8 bytes on 64bit platform and 4 bytes on 32bit platform.
        if (CPreferenceStore.is64bitAddress()) {
            addr_type.setWidth(8);
            addr_type.setBitModePad(AddressMode.MODE64_LITERAL);
            setBaseTDTypeAlignment(addr_type, 8);
        } else {
            addr_type.setWidth(4);
            addr_type.setBitModePad(AddressMode.MODE32_LITERAL);
            setBaseTDTypeAlignment(addr_type, 4);
        }

        addr_type.setBigEndian(new Boolean(CPreferenceStore.isBigEndian()));
        addr_type.setAbsolute(Boolean.TRUE);

        return addr_type;
    }

    /**
     * Return the PlatformCompilerInfo object. Do the initialize if necessary.
     */
    public static PlatformCompilerInfo getPlatformCompilerInfo() {
        if (_platformCompilerInfo == null)
            createPlatformCompilerInfo();
        return _platformCompilerInfo;
    }

    /**
     * Create a TypeDescriptor type for int
     */
    public static IntegerTD createInt() {
        return createIntegralType(4, true);
    }

    /**
     * Create a TypeDescriptor type for unsigned int
     */
    public static IntegerTD createUnsignedInt() {
        return createIntegralType(4, false);
    }

    /**
     * Create a TypeDescriptor type for char. Char is represented by a StringTD
     * object.
     */
    public static StringTD createChar() {
        return createStringType(1, 1, true);
    }

    /**
     * Create a TypeDescriptor type for unsigned char. Unsigned char is
     * represented by an integer rather than a string.
     */
    public static IntegerTD createUnsignedChar() {
        return createIntegralType(1, false);
    }

    /**
     * Create a TypeDescriptor type for signed char. Signed char is also
     * represented by an integer.
     */
    public static IntegerTD createSignedChar() {
        return createIntegralType(1, true);
    }

    /**
     * Create a TypeDescriptor type for short
     */
    public static IntegerTD createShort() {
        return createIntegralType(2, true);
    }

    /**
     * Create a TypeDescriptor type for unsigned short
     */
    public static IntegerTD createUnsignedShort() {
        return createIntegralType(2, false);
    }

    /**
     * Create a TypeDescriptor type for long
     */
    public static IntegerTD createLong() {
        if (CPreferenceStore.is64bitAddress())
            return createIntegralType(8, true);
        else
            return createIntegralType(4, true);
    }

    /**
     * Create a TypeDescriptor type for unsigned long
     */
    public static IntegerTD createUnsignedLong() {
        if (CPreferenceStore.is64bitAddress())
            return createIntegralType(8, false);
        else
            return createIntegralType(4, false);
    }

    /**
     * Create a TypeDescriptor type for long long
     */
    public static IntegerTD createLongLong() {
        return createIntegralType(8, true);
    }

    /**
     * Create a TypeDescriptor type for unsigned long long
     */
    public static IntegerTD createUnsignedLongLong() {
        return createIntegralType(8, false);
    }

    /**
     * Calculate Enumeration size Retrieve the enumeration size from the data
     * store. If the user wants enums 'compacted' then we need to figure out the
     * max value of the enums and the storage it will fit into.
     */
    public static int calculateEnumSize(CEnumeration cEnum) {
        int enumSize = CPreferenceStore.getSizeEnums();
        // if we are compacting enumerations - figure how many bytes it has to
        // be
        if (enumSize == SIZE_OF_ENUM_COMPACT_VALUE) { // @P3
            int maxValue = 0;
            int minValue = 0;
            boolean fMax = true;
            /* Keep track of the max and min values for all enumerators */
            /* If we get a new max value and the following enum doesn't */
            /* have a value then the max value is incremented. */
            org.eclipse.emf.common.util.EList enums = cEnum.getEnumerator();
            for (int i = 0; i < enums.size(); i++) {
                CEnumerator enum = (CEnumerator) enums.get(i);
                if (enum.getValue() != null) {
                    int value = enum.getValue().intValue();
                    if (value > maxValue) {
                        maxValue = value;
                        fMax = true;
                    } else {
                        fMax = false;
                        if (value < minValue)
                            minValue = value;
                    }
                } else if (fMax)
                    maxValue++;
            }

            if (maxValue <= 127 && minValue >= -128)
                enumSize = 1;
            else if (maxValue <= 32767 && minValue >= -32768)
                enumSize = 2;
            else
                enumSize = 4;
        } else if (enumSize != 1 && enumSize != 2 && enumSize != 4)
            enumSize = 4;
        return enumSize;
    }

    /**
     * Create a TypeDescriptor type for an enum Enum is normally represented by
     * a 4 byte signed integer.
     */
    public static IntegerTD createEnumeration(CEnumeration cEnum) {
        return createIntegralType(calculateEnumSize(cEnum), true);
    }

    /**
     * Create a TypeDescriptor type for wchar_t. wchar_t is represented by a
     * StringTD object.
     */
    public static StringTD createWchar() {
        int size = CPreferenceStore.getSizeWchar(); //@P4A
        return createStringType(size, size, false); //@P4C
    }

    /**
     * Create a TypeDescriptor type for float
     */
    public static FloatTD createFloat() {
        return createFloatingType(4);
    }

    /**
     * Create a TypeDescriptor type for double
     */
    public static FloatTD createDouble() {
        return createFloatingType(8);
    }

    /**
     * Create a TypeDescriptor type for long double
     */
    public static FloatTD createLongDouble() {
        if (CPreferenceStore.isLongDouble128())
            return createFloatingType(16);
        else
            return createFloatingType(8);
    }

    /**
     * Set the offset and size for field instances within a struct/union,
     * including the fields within the nested structs/unions.
     */
    public static void setInstancesOffsetAndSize(CStructured astructured) throws CException {
        CStructuredDetails details = CTypeDescriptorHelper.getStructuredDetails(astructured);
        List fields = astructured.getContains();
        CStructured container = astructured.getContainer();

        // If this struct/union is contained in another struct/union, the base
        // offset will
        // be added to the offset of each field instance.
        int baseOffset = 0;
        if (container != null)
            baseOffset = CTypeDescriptorHelper.offsetOf(astructured);

        int i = -1;
        Iterator it = fields.iterator();
        while (it.hasNext()) {
            Object nextObject = it.next();
            i++;
            if (nextObject instanceof CField) {
                CField afield = (CField) nextObject;
                InstanceTDBase instance = afield.getInstanceTDBase();
                if (instance.getArrayDescr().size() == 0) {
                    instance.setSize(String.valueOf(details.getFieldSize(i)));
                    instance.setContentSize(instance.getSize());
                }
                instance.setOffset(String.valueOf(baseOffset + details.getFieldOffset(i)));
            } else if (nextObject instanceof CStructured) {
                CStructured nestedStructured = (CStructured) nextObject;
                setInstancesOffsetAndSize(nestedStructured);
            }
        }
    }

    /**
     * Set the alignment of a BaseTDType. The alignment is the minimum of the
     * width and the pack level.
     */
    private static void setBaseTDTypeAlignment(BaseTDType tdType, int width) {
        int alignment = Math.min(width, CPreferenceStore.getPackLevel());
        switch (alignment) {
        case 1:
            tdType.setAlignment(AlignType.BYTE_LITERAL);
            break;
        case 2:
            tdType.setAlignment(AlignType.HALFWORD_LITERAL);
            break;
        case 4:
            tdType.setAlignment(AlignType.WORD_LITERAL);
            break;
        case 8:
        default:
            tdType.setAlignment(AlignType.DOUBLEWORD_LITERAL);
            break;
        }
    }

    /**
     * Map the float type string to an enum value
     */
    private static FloatValue mapFloatValue(String floatType) {
        if (floatType == null)
            return FloatValue.UNSPECIFIED_LITERAL;
        else if (floatType.equals(IEEE_EXTENDED_INTEL))
            return FloatValue.IEEE_EXTENDED_INTEL_LITERAL;
        else if (floatType.equals(IEEE_EXTENDED_AIX))
            return FloatValue.IEEE_EXTENDED_AIX_LITERAL;
        else if (floatType.equals(IEEE_EXTENDED_OS390))
            return FloatValue.IEEE_EXTENDED_OS390_LITERAL;
        else if (floatType.equals(IEEE_EXTENDED_AS400))
            return FloatValue.IEEE_EXTENDED_AS400_LITERAL;
        else if (floatType.equals(IEEE_NON_EXTENDED))
            return FloatValue.IEEE_NON_EXTENDED_LITERAL;
        else if (floatType.equals(IBM_390_HEX))
            return FloatValue.IBM390_HEX_LITERAL;
        else if (floatType.equals(IBM_400_HEX))
            return FloatValue.IBM400_HEX_LITERAL;
        else
            return FloatValue.UNSPECIFIED_LITERAL;
    }

    /**
     * Return the TypeDescriptor type for a C model non-aggregate type
     * 
     * @return The corresponding Type Descriptor type
     * @param type
     *            A C model type
     */
    public static BaseTDType getTDType(CClassifier type) throws Exception {

        BaseTDType returnType = null;

        // A char or wchar_t array is presented by a StringTD object.
        if (CQuery.isWcharArray(type) || (CQuery.isUnsignedShortArray(type) && CPreferenceStore.isTreatUShortArrayAsDBCS())) { //@P1A
            BaseTDType tdType = CPrimitiveTypesMapper.getTDTypeFromCType(type);
            if (tdType != null)
                return tdType;
            else {
                int size; //@P4A {
                if (CQuery.isWcharArray(type))
                    size = CPreferenceStore.getSizeWchar();
                else
                    size = 2; //@P4A }
                int width = size * ((CArray) type).getDimension().intValue(); //@P4C
                StringTD stringType = createStringType(width, size, false);
                CPrimitiveTypesMapper.putTDType(type, stringType);
                return stringType;
            }
        } else if (CQuery.isCharArray(type)) {
            BaseTDType tdType = CPrimitiveTypesMapper.getTDTypeFromCType(type);
            if (tdType != null)
                return tdType;
            else {
                int width = ((CArray) type).getDimension().intValue();
                StringTD stringType = createStringType(width, 1, false);
                CPrimitiveTypesMapper.putTDType(type, stringType);
                return stringType;
            }
        } else {

            if (CQuery.isArray(type)) {
                Object derivable = ((CArray) type).getDerives();
                if (derivable instanceof CClassifier)
                    type = (CClassifier) derivable;
            }
            // A pointer is represented by an AddressTD object
            if (type instanceof CPointer) {
                returnType = createAddressForPointer((CPointer) type);
            } else if (type instanceof CDatatype) {
                // Is it an enumeration?
                if (type instanceof CEnumeration) {
                    returnType = createEnumeration((CEnumeration) type);
                }
                // Is it a bitfield?
                else if (type instanceof CBitField) {
                    BaseTDType tdType = CPrimitiveTypesMapper.getTDTypeFromCType(type);
                    if (tdType != null) {
                        returnType = tdType;
                    } else {
                        int size = ((CBitField) type).getSize().intValue();
                        returnType = createBinaryTypeForBitField(size);
                        CPrimitiveTypesMapper.putTDType(type, returnType);
                    }
                }
                // Other primitive types are directly retrieved from the mapper.
                else
                    returnType = CPrimitiveTypesMapper.getTDTypeFromCType((CDatatype) type);
            } else {
                throw new CException(CResource.getString("_ERROR_TypeDescriptor_Type_Invalid", type.getName()));
            }
        }
        return returnType;
    }

    /**
     * Create a TypeDescriptor AddressTD object for a C model pointer type
     * 
     * @return The corresponding AddressTD object for the given CPointer object
     * @param apointer
     *            A pointer type in the C model
     */
    private static AddressTD createAddressForPointer(CPointer apointer) {
        // If the TD type already exists in the TDMap, just return it.
        BaseTDType tdType = CPrimitiveTypesMapper.getTDTypeFromCType(apointer);
        if (tdType != null && tdType instanceof AddressTD)
            return (AddressTD) tdType;

        AddressTD addressType = createAddressType();
        CDerivableType referenceType = apointer.getDerives();
        if (referenceType instanceof CClassifier) {
            CClassifier type = (CClassifier) referenceType;

            if (type instanceof CTypedef)
                type = CQuery.typedefBaseType((CTypedef) type);

            if (CQuery.isPointer(type) || CQuery.isPrimitiveType(type)) {
                BaseTDType baseTdType = null;

                try {
                    baseTdType = getTDType(type);
                } catch (Exception e) {
                }

                if (baseTdType != null) {
                    addressType.setReferenceType(baseTdType);
                }
            }
        }

        // Save the mapping in the TDMap
        CPrimitiveTypesMapper.putTDType(apointer, addressType);

        return addressType;
    }

}